# copyright 2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-trackervcs views"""

from cubicweb.view import EntityView
from cubicweb.selectors import is_instance, score_entity, adaptable
from cubicweb.web import uicfg, facet
from cubicweb.web.views import ibreadcrumbs


# XXX in cw > 3.12.4 as 'incontext-state'
class InContextWithStateView(EntityView):
    __regid__ = 'myincontext-state'
    __select__ = adaptable('IWorkflowable')
    def entity_call(self, entity):
        iwf = entity.cw_adapt_to('IWorkflowable')
        self.w(u'%s [%s]' % (entity.view('incontext'), iwf.printable_state))


class TicketInContextWithStateAndVersionView(EntityView):
    __regid__ = 'incontext-state-version'
    __select__ = is_instance('Ticket')
    def entity_call(self, entity):
        state = entity.cw_adapt_to('IWorkflowable').printable_state
        if entity.done_in:
            self.w(u'%s [%s, %s]' % (entity.view('incontext'), state,
                                     entity.done_in[0].view('incontext')))
        else:
            self.w(u'%s [%s]' % (entity.view('incontext'), state))


class ProjectPatchesTab(EntityView):
    __regid__ = 'vcreview.patches_tab'
    __select__ = is_instance('Project') & score_entity(
        lambda x: x.source_repository and x.source_repository[0].has_review)

    def entity_call(self, entity):
        entity.source_repository[0].view(
            'vcreview.repository.patches', w=self.w)


class VersionPatchesTab(EntityView):
    __regid__ = 'trackervcs.version_patches_tab'
    __select__ = is_instance('Version') & score_entity(
        lambda x: x.project.source_repository and x.project.source_repository[0].has_review)

    def entity_call(self, entity):
        rql = ('Any P,P,P,P,PB,PS GROUPBY P,PB,PS '
               'WHERE P patch_revision RE, RE branch PB, P in_state PS, '
               'RE from_repository R, P patch_ticket T, '
               'T done_in V, V eid %(v)s')
        rset = self._cw.execute(rql, {'v': entity.eid})
        self.wview('vcreview.patches.table', rset, 'noresult', displayfilter=True)

class HasPatchFacet(facet.HasRelationFacet):
    __regid__ = 'trackervcs.has-patch-facet'
    rtype = 'patch_ticket'
    role = 'object'

class HasTicketFacet(facet.HasRelationFacet):
    __regid__ = 'trackervcs.has-ticket-facet'
    rtype = 'patch_ticket'
    role = 'subject'


def registration_callback(vreg):
    if not 'vcreview' in vreg.config.cubes():
        return # don't register anything from this module

    from cubicweb.web.views import tableview
    from cubes.vcreview.views import (startup as vcreview_startup,
                                      primary as vcreview_primary)
    from cubes.tracker.views import project, version

    vreg.register_all(globals().values(), __name__)

    # override vcreview summary tables to include patch's ticket and ticket's
    # version
    vcreview_startup.AllActivePatches.rql = (
        'Any P,P,P,P,PB,PS,R,T,V GROUPBY R,P,PB,PS,T,V '
        'ORDERBY RT,PB WHERE P patch_revision RE, RE branch PB, P in_state PS, '
        'RE from_repository R, R title RT, '
        'NOT PS name %s, '
        'P patch_ticket T?, T done_in V?')
    vcreview_startup.UserWorkList.rql = (
        'Any P,P,P,P,PB,PS,R,T,V GROUPBY R,P,PB,PS,T,V '
        'ORDERBY RT,PB WHERE P patch_revision RE, RE branch PB, P in_state PS,'
        'RE from_repository R, R title RT, U eid %(u)s, '
        'EXISTS(PS name "pending-review" AND P patch_reviewer U) '
        'OR EXISTS(PS name "reviewed" AND P patch_committer U) '
        'OR (EXISTS(PS name IN ("reviewed", "deleted") AND NOT EXISTS(P patch_committer U) '
        '    AND (EXISTS(R repository_committer U) OR EXISTS(U in_group G, G name "committers")))), '
        'P patch_ticket T?, T done_in V?')
    vcreview_primary.RepositoryPatchesTable.rql = (
        'Any P,P,P,PB,PS,T,V GROUPBY P,PB,PS,T,V WHERE '
        'P patch_revision RE, RE branch PB, P in_state PS, '
        'RE from_repository R, R eid %(x)s, '
        'P patch_ticket T?, T done_in V?')
    # add a tab on project view to display patches
    project.ProjectPrimaryView.tabs.append(ProjectPatchesTab.__regid__)
    # uicfg tweaks
    _pvs = uicfg.primaryview_section
    _pvdc = uicfg.primaryview_display_ctrl
    _pvs.tag_subject_of(('Patch', 'patch_ticket', '*'), 'attributes')
    _pvdc.tag_subject_of(('Patch', 'patch_ticket', '*'), {'vid': 'incontext-state-version'})
    _pvs.tag_object_of(('*', 'patch_ticket', 'Ticket'), 'attributes')
    _pvdc.tag_object_of(('*', 'patch_ticket', 'Ticket'), {'vid': 'myincontext-state'})

    project.ProjectTicketsTable.columns.append('patch_ticket')
    project.ProjectTicketsTable.column_renderers['patch_ticket'] = \
        tableview.RelationColRenderer(role='object')
    version.VersionTicketsTable.columns.append('patch_ticket')
    version.VersionTicketsTable.column_renderers['patch_ticket'] = \
        tableview.RelationColRenderer(role='object')
    version.VersionPrimaryView.tabs.append(_('trackervcs.version_patches_tab'))




pwf = rql('Workflow X WHERE X name "patch workflow"').get_entity(0, 0)
for trname in ('ask review', 'reject', 'fold'):
    tr = pwf.transition_by_name(trname)
    rql('DELETE X require_group G WHERE X eid %(x)s, G name "staff"', {'x': tr.eid})



# copyright 2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-trackervcs views"""

_ = unicode

from cubicweb.view import EntityView
from cubicweb.selectors import is_instance, score_entity
from cubicweb.web import uicfg
from cubicweb.web.views import ibreadcrumbs

from cubes.tracker.views import project

_pvs = uicfg.primaryview_section
_pvdc = uicfg.primaryview_display_ctrl
# _abaa = uicfg.actionbox_appearsin_addmenu
# _afs = uicfg.autoform_section
# _affk = uicfg.autoform_field_kwargs

_pvs.tag_subject_of(('Project', 'source_repository', '*'), 'attributes')
_pvs.tag_subject_of(('Ticket', 'closed_by', '*'), 'attributes')


class SourceRepositoryIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance('Repository') & score_entity(lambda x: x.reverse_source_repository)

    def parent_entity(self):
        return self.entity.reverse_source_repository[0]


class ProjectRepoTab(EntityView):
    __regid__ = _('vcreview.project.vcsrepo_tab')
    __select__ = is_instance('Project') & score_entity(
        lambda x: x.source_repository and x.source_repository[0].source_url)

    def entity_call(self, entity):
        entity.source_repository[0].view(
            'vcsfile.repository.checkout', w=self.w)

project.ProjectPrimaryView.tabs.append(ProjectRepoTab.__regid__)




options = (
    ('trusted-vcs-repositories',
     {'type' : 'yn',
      'default': True,
      'help': 'Are repositories trusted, in which case checkin messages are '
      'analyzed to trigger some actions such as changing ticket\'states.',
      'level': 2, 'group': 'vcsfile',
      }),
    )




for etype in ('Patch', 'Task', 'VersionContent', 'InsertionPoint'):
    if not (etype, 'CWPermission') in schema['require_permission'].rdefs:
        add_relation_definition(etype, 'require_permission', 'CWPermission')

sync_schema_props_perms('Patch', syncprops=False)
sync_schema_props_perms('Task', syncprops=False)



def registration_callback(vreg):
    if 'vcreview' not in vreg.config.cubes():
        return

    from cubes.vcreview import sobjects as vcreview

    class PatchStatusChangeView(vcreview.PatchStatusChangeView):
        def subject(self):
            patch = self.cw_rset.get_entity(0, 0)
            if patch.repository.project:
                repo_title = patch.repository.project.dc_title()
            else:
                repo_title = patch.repository.dc_title()
            return self._cw.__(u'[%(repo)s] patch %(state)s: %(patch)s') % {
                'repo': repo_title,
                'patch': patch.dc_title(),
                'state': self._cw.__(self._kwargs['current_state'])}

    class TaskAddedView(vcreview.TaskAddedView):
        def subject(self):
            task = self.cw_rset.get_entity(0, 0)
            patch = task.activity_of
            if not patch:
                return super(TaskAddedView, self).subject()
            if patch.__regid__ == 'InsertionPoint':
                patch = patch.parent.reverse_patch_revision
                if not patch:
                    return super(TaskAddedView, self).subject()
                patch = patch[0]
            if patch.repository.project:
                repo_title = patch.repository.project.dc_title()
            else:
                repo_title = patch.repository.dc_title()
            return self._cw.__(u'[%(repo)s] patch task: %(patch)s - %(task)s') % {
                'repo': repo_title,
                'patch': patch.dc_title(),
                'task': task.dc_title()}

    class ReviewerSetNotificationView(vcreview.ReviewerSetNotificationView):
        def subject(self):
            patch = self.cw_rset.get_entity(0, 0)
            if patch.repository.project:
                repo_title = patch.repository.project.dc_title()
            else:
                repo_title = patch.repository.dc_title()
            return self._cw.__('[%(repo)s] patch assigned: %(patch)s') % {
                'repo': repo_title,
                'patch': patch.dc_title()}

    vreg.register_and_replace(PatchStatusChangeView, vcreview.PatchStatusChangeView)
    vreg.register_and_replace(TaskAddedView, vcreview.TaskAddedView)
    vreg.register_and_replace(ReviewerSetNotificationView, vcreview.ReviewerSetNotificationView)



# copyright 2011-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-trackervcs schema"""

from yams.buildobjs import RelationDefinition
from yams.reader import context

from cubicweb.schema import RQLConstraint


class source_repository(RelationDefinition):
    """version controled system holding sources for this project"""
    subject = 'Project'
    object = 'Repository'
    cardinality = '??'


class introduced_by(RelationDefinition):
    subject = 'Ticket'
    object = 'Revision'
    constraints = [RQLConstraint('S concerns P, O from_repository R, '
                                 'EXISTS(P source_repository R) OR EXISTS(P subproject_of PP, PP source_repository R)')]


class closed_by(RelationDefinition):
    subject = 'Ticket'
    object = 'Revision'
    constraints = [RQLConstraint('S concerns P, O from_repository R, '
                                 'EXISTS(P source_repository R) OR EXISTS(P subproject_of PP, PP source_repository R)')]


class require_permission(RelationDefinition):
    subject = ('VersionedFile', 'Revision',
               'DeletedVersionContent', 'VersionContent')
    object = 'CWPermission'


if 'Patch' in context.defined:
    from cubes.vcreview.schema import Patch
    from cubes.task.schema import Task

    class patch_ticket(RelationDefinition):
        subject = 'Patch'
        object = 'Ticket'
        constraints = [RQLConstraint('S patch_revision RE, '
                                     'RE from_repository R, '
                                     'O concerns P, '
                                     'EXISTS(P source_repository R) OR EXISTS(P subproject_of PP, PP source_repository R)')]


    require_permission.subject += ('Patch', 'Task', 'InsertionPoint')

    Task.__permissions__.update({'add': ('managers', 'staff',),
                                 'update': ('managers', 'staff', 'owners')})
    Patch.__permissions__.update({'update': ('managers', 'staff')})



from logilab.common.decorators import monkeypatch
from cubes.vcsfile import entities as vcsfile
from cubes.vcreview import entities as vcsreview

@monkeypatch(vcsfile.Repository, 'project')
@property
def project(self):
    if self.reverse_source_repository:
        return self.reverse_source_repository[0]
    return None

class PatchReviewBehaviour(vcsreview.PatchReviewBehaviour):
    """Select a reviewer for a patch.
    Try first to filter cwusers to ones that are already reviewers of
    a patch related to the same ticket. If the filter returns an empty
    set we fallback to the previous pre-selection.
    """
    def reviewers_rset(self):
        rset = self._cw.execute(
            'DISTINCT Any U, "" '
            'WHERE P eid %(p)s, '
            '      P patch_ticket T, '
            '      PS patch_ticket T, '
            '      PS patch_reviewer U',
            {'p': self.entity.eid})
        if not rset:
            rset = super(PatchReviewBehaviour, self).reviewers_rset()
        return rset

def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__, butclasses=(PatchReviewBehaviour,))
    vreg.register_and_replace(PatchReviewBehaviour, vcsreview.PatchReviewBehaviour)




for etype in ('VersionedFile', 'Revision',
              'DeletedVersionContent', 'VersionContent'):
    # may be already in. Or not...
    add_relation_definition(etype, 'require_permission', 'CWPermission')

rql('SET X require_permission P WHERE '
    'R require_permission P, X patch_revision R, '
    'NOT X require_permission P')
commit()



"""trackervcs automatic tests"""

from cubicweb.devtools.fill import ValueGenerator
from cubicweb.devtools.testlib import AutomaticWebTest

class MyValueGenerator(ValueGenerator):
    def generate_Repository_path(self, entity, index):
        return u'/' + self.generate_string(entity, 'path', index)

    def generate_Project_icon(self, entity, index, **kwargs):
        return None

class AutomaticWebTest(AutomaticWebTest):
    '''provides `to_test_etypes` and/or `list_startup_views` implementation
    to limit test scope
    '''
    no_auto_populate = ('VersionContent', 'DeletedVersionContent',
                        'Revision', 'InsertionPoint', 'Patch', 'VersionedFile')
    ignored_relations = set(('nosy_list', 'subproject_of'))

    def to_test_etypes(self):
        '''only test views for entities of the returned types'''
        return set()

    def list_startup_views(self):
        '''only test startup views of the returned identifiers'''
        return ('vcreview.allactivepatches', 'vcreview.patches.worklist')

if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2011-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-trackervcs specific hooks and operations"""

import re

from rql import TypeResolverException

from cubicweb import typed_eid
from cubicweb.server import hook
from cubicweb.selectors import is_instance

from cubes.tracker import hooks as tracker

_BASE_RE = r'\b%s\b\s+(#\d+(?:(?:,|\s|\s+and\s)\s*#\d+)*)'
CLOSES_EID_RGX = re.compile(_BASE_RE % 'closes', re.IGNORECASE)
RELATED_EID_RGX = re.compile(_BASE_RE % '(?:closes|related to)', re.IGNORECASE)
TICKET_EID_RGX = re.compile(r'#(\d+)')

def closed_eids(line):
    """Parse closes command in description and return target eids"""
    eids = []
    for cmd in CLOSES_EID_RGX.findall(line):
        eids.extend(int(m) for m in TICKET_EID_RGX.findall(cmd))
    return eids

def related_eids(line):
    """Parse closes and related commands in description and return target eids
    """
    eids = []
    for cmd in RELATED_EID_RGX.findall(line):
        eids.extend(int(m) for m in TICKET_EID_RGX.findall(cmd))
    return eids

def tickets(session, ticket_eids, repository_eid):
    """Return rset for a ticket referenced in a commit message.

    The ticket must come from a project using given repository as
    source_repository (or one of its subproject).

    Return None if no match.
    """
    for ticket_eid in ticket_eids:
        try:
            rset = session.execute('Any X, P WHERE X eid %(x)s, X concerns P',
                                   {'x': ticket_eid})
        except TypeResolverException: # ticket_eid is not a ticket
            continue
        if not rset: # ticket_eid doesn't exist
            continue
        project = rset.get_entity(0, 1)
        while project:
            if project.source_repository:
                if project.source_repository[0].eid == repository_eid:
                    yield rset.get_entity(0, 0)
                else:
                    break
            project = project.subproject_of and project.subproject_of[0]

# search for ticket closing instructions #######################################


_CLOSED_BY_RQL = '''
SET T closed_by R
WHERE NOT T closed_by R,
      T eid %(t)s,
      R eid %(r)s
'''
class SearchRevisionTicketOp(hook.DataOperationMixIn,
                             hook.Operation):
    """Search magic words in revision's commit message:

        closes #<ticket eid>

    When found, the ticket is marked as closed by the revision and its state is
    changed.
    """
    def precommit_event(self):
        for rev_eid in self.get_data():
            # search for instruction
            rev = self.session.entity_from_eid(rev_eid)
            eids = closed_eids(rev.description)
            for ticket in tickets(self.session, eids, rev.repository.eid):
                if rev.phase == 'public':
                    self.session.execute(_CLOSED_BY_RQL,
                                         {'t': ticket.eid, 'r': rev.eid})
                    iwf = ticket.cw_adapt_to('IWorkflowable')
                    for tr in iwf.possible_transitions():
                        if tr.name in ('done', 'close'):
                            iwf.fire_transition(tr)
                            break

class RevisionAdded(hook.Hook):
    """Detect that a new revision is touched to prepare relevant operation

    On update we ignore non-phase change.
    """
    __regid__ = 'trackervcs.revision-to-ticket-state'
    __select__ = hook.Hook.__select__ & is_instance('Revision')
    events = ('after_add_entity', 'after_update_entity')
    category = 'autoset'

    def __call__(self):
        repo = self.entity.repository
        if not repo.reverse_source_repository:
            return
        if 'update' in self.event and 'phase' not in self.entity.cw_edited:
            return
        SearchRevisionTicketOp.get_instance(self._cw).add_data(self.entity.eid)


# synchronize  "P source_repository R" with ####################################
# "P has_apycot_environment PE AND PE local_repository R"

class SetSourceRepositoryOp(hook.Operation):
    """Automatically assign project repo to test env repo"""
    def precommit_event(self):
        if not self.project.source_repository and self.projectenv.repository:
            repo = self.projectenv.repository
            self.project.set_relations(source_repository=repo)

class ApycotEnvironmentAdded(hook.Hook):
    """Automatically assign project repo to test env repo"""
    __regid__ = 'trackervcs.add-source-repository'
    __select__ = hook.Hook.__select__ & hook.match_rtype('has_apycot_environment')
    events = ('after_add_relation',)
    category = 'autoset'

    def __call__(self):
        project = self._cw.entity_from_eid(self.eidfrom)
        if not project.source_repository:
            projectenv = self._cw.entity_from_eid(self.eidto)
            SetSourceRepositoryOp(self._cw, project=project, projectenv=projectenv)

# set ticket 'in-progress' when attached to a patch ############################

class PatchLinkedToTicket(hook.Hook):
    """Set ticket 'in-progress' when attached to a patch"""
    __regid__ = 'trackervcs.patch-linked-to-ticket'
    __select__ = hook.Hook.__select__ & hook.match_rtype('patch_ticket')
    events = ('after_add_relation',)
    category = 'autoset'

    def __call__(self):
        ticket = self._cw.entity_from_eid(self.eidto)
        ticket.cw_adapt_to('IWorkflowable').fire_transition_if_possible('start')

# attach patch to ticket on addition ###########################################

class PatchRevisionAdded(hook.Hook):
    """Attach patch to ticket on addition"""
    __regid__ = 'trackervcs.patch-revision-added'
    __select__ = hook.Hook.__select__ & hook.match_rtype('patch_revision')
    events = ('after_add_relation',)
    category = 'autoset'

    def __call__(self):
        patch = self._cw.entity_from_eid(self.eidfrom)
        rev = self._cw.entity_from_eid(self.eidto)
        source_repo = rev.repository
        eids = related_eids(rev.description)
        for ticket in tickets(self._cw, eids, source_repo.eid):
            # XXX should we drop existing links first?
            # We could just run the RQL below if there wasn't a cache issue
            # SET P patch_ticket T WHERE NOT P patch_ticket T, P eid ...
            if all(x.eid != ticket.eid for x in patch.patch_ticket):
                patch.set_relations(patch_ticket=ticket)


# security propagation #########################################################

# take care, code below may incidentally be propagated to nosylist.hooks.S_RELS
# depending on trackervcs/forge import order. Import order is fixed using
# __pkginfo__.__recommends__ so everything should be ok
tracker.S_RELS.add('source_repository')
tracker.O_RELS |= set(('from_repository', 'from_revision'))

# registration control #########################################################

def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__, (RevisionAdded,
                                                     PatchLinkedToTicket,
                                                     PatchRevisionAdded,
                                                     ApycotEnvironmentAdded))
    if vreg.config['trusted-vcs-repositories']:
        vreg.register(RevisionAdded)

    if 'vcreview' in vreg.config.cubes():
        vreg.register(PatchLinkedToTicket)
        vreg.register(PatchRevisionAdded)
        # vcreview permission propagation ######################################
        tracker.S_RELS |= set(('has_activity', ))
        tracker.O_RELS |= set(('patch_revision', 'point_of',))
    if 'apycot' in vreg.config.cubes():
        vreg.register(ApycotEnvironmentAdded)
